import subprocess


LETTERS = list('abcdefghijklmnopqrstuvwxyz')
for _letter in 'abcd':
    LETTERS.extend(_letter + c for c in 'abcdefghijklmnopqrstuvwxyz')


class Sheet:
    """
    Represents a spreadsheet running in a LOCalc instance.
    """
    
    def __init__(self, file):
        self.file = file
        self.process = None

    def __del__(self):
        self.kill()

    def kill(self):
        "Kill current libreoffice process"
        self.process.kill()
        self.process = None

    def start(self):
        "Start calc process"
        cmd = [
            'libreoffice', '--calc',
            '--accept="socket,host=localhost,port=2002;urp;StarOffice.ServiceManager"',
            self.file,
        ]
        self.process = subprocess.Popen(cmd)
        return self.process

    def sheet(self):
        """
        Return an object representing the spreadsheet.
        
        Must start libreoffice process first.
        """
        if self.process is None:
            raise RuntimeError('must start libreoffice before starting.')

        from oosheet import OOSheet
        return OOSheet

    def template_cells(self):
        """
        Search all cells for jinja2 templates
        """
        S = self.sheet()
        ROWS = 57
        COLS = LETTERS.index('m')
        cell_map = {}
        empty_cols = 0
        empty_rows = 0

        for i in range(ROWS):
            for j in range(COLS):
                cell_name = cell(i, j)
                data = S(cell_name).string.strip()
                if data.startswith('{') and data.endswith('}'):
                    cell_map[cell_name] = data
            
        return cell_map


def cell(i, j):
    "Convert coordinates i,j to cell names"
    return LETTERS[j] + str(i + 1)



import click
import os
from . import db, sheets, templates 


@click.group(name='buroca')
def buroca():
    "A small utility for automatic generation of paperwork."


@buroca.command()
def init():
    "create initial paths."

    abspath = (lambda x: os.path.join(curr_path, x))
    curr_path = os.getcwd()

    for path in ['data', 'reports', 'templates']:
        path_ = abspath(path)
        if not os.path.exists(path_):
            click.echo('creating %r...' % path)
            os.mkdir(path_)

    click.echo('Paths successfully created!')


@buroca.command()
@click.argument('template')
@click.argument('resource')
def do(template, resource):
    "create reports from resources and templates."

    template_base, template_ext = os.path.splitext(template)

    # Grab template
    template_path = project_path('templates', template)
    if not os.path.exists(template_path):
        raise SystemExit('template does not exist: %s' % template)
    
    jinja_template = templates.load_template(template_path)
    
    # Load resources
    if resource.endswith('/*'):
        resources = db.find_resources(project_path(), resource[:-2], True)
    else:
        resource_name, entity = resource.split('/')
        resources = {entity: db.load_resources(project_path(), entity, True)}
    
    # Apply template
    for entity, ns in resources.items():
        data = jinja_template.render(ns)
        fname = '%s-%s%s' % (template_base, entity, template_ext)
        
        with open(project_path('reports', fname), 'w') as F:
            F.write(data)
            click.echo('Report: reports/%s.' % fname)


def project_path(*args):
    "Return a subpath into the project's tree"

    return os.path.join(os.getcwd(), *args)



import sys

from setuptools import setup, find_packages

sys.path.append('src')

setup(
    package_dir={'': 'src'},
    packages=find_packages('src'),
    setup_requires = ['setuptools >= 30.3.0'],
)



"""
Load and locate YAML resources from the /data/ folder of a project.
"""

import os
from types import SimpleNamespace
from functools import lru_cache

import yaml


#
# Load YAML resources.
#
def load_resources(base, name, as_namespace=False):
    """
    Load all resources on the given path.

    Args:
        base (str):
            Project directory. Must contain a /data/ folder with yaml files. 
        name (str):
            Name of the required resource. It will scan this resource on all
            sub-folders and associate the folder name with folder/<name>.yml
            in the resulting namespace.
        as_namespace (bool):
            If True, return each resource as a SimpleNamespace instance instead
            of a a dictionary.

    Return:
        A dictionary mapping resource names to values.
    """

    datadir = os.path.join(base, 'data')
    paths = os.listdir(datadir)
    ns = {}

    for path in paths:
        fullpath = os.path.join(datadir, path)

        if is_resource(path):
            ns[path[:-4]] = load_yaml(fullpath)

        elif os.path.isdir(fullpath):
            resource_path = os.path.join(fullpath, name) + '.yml'
            resource_name = os.path.split(fullpath)[-1]
            data = (load_yaml(resource_path)
                    if os.path.exists(resource_path) else None)
            ns[resource_name] = data

    if as_namespace:
        for name, value in ns.items():
            if isinstance(name, dict):
                ns[name] = SimpleNamespace(**value)
    return ns


def find_resources(base, path_from=None, as_namespace=False):
    """
    Return a mapping of entity to namespace using the given path as reference
    to locate entity values.
    """

    path_from = path_from or locate_entities(base)
    resource_path = os.path.join(base, 'data', path_from)
    files = os.listdir(resource_path)
    names = [f[:-4] for f in files if is_resource(f)]
    return {name: load_resources(base, name, as_namespace) for name in names}


def locate_entities(base):
    """
    Return a possible reference path for a directory that stores project's
    entities
    """
    datadir = os.path.join(base, 'data')
    paths = (os.path.join(datadir, f) for f in os.listdir(datadir))
    folders = [f for f in paths if os.path.isdir(f)]

    if not folders:
        return None
    elif len(folders) == 1:
        name, = folders
        return os.path.split(name)[-1]
    else:
        files = {f: {f for f in os.listdir(f) if is_resource(f)}
                 for f in folders}
        resource_set = set()
        resource_path = None

        for path, data in files.items():
            if resource_set.issubset(data):
                resource_set = data
                resource_path = path
            elif resource_set.issuperset(data):
                pass
            else:
                msg = 'inconsistent resources: %s and %s'
                raise TypeError(msg % (resource_path, path))
        
        return resource_path


@lru_cache(1024)
def load_yaml(path):
    "Load YAML and return as a simple namespace"
    return yaml.safe_load(open(path))


def is_resource(file):
    "Return True if file extension indicates it is a db resource."
    return file.endswith('.yml')



import subprocess


def headless_cmd(*args):
    """
    Execute libreoffice in headless mode
    """
    cmd = ['libreoffice', '-headless']
    subprocess.run(cmd)


def odt_to_rtf(source, dest):
    "Convert .odt file to rtf"

    return headless_cmd('-convert-to', 'rtf', source, '-o', dest)


def odt_to_pdf(source, dest):
    "Convert .odt to pdf"

    return headless_cmd('-convert-to', 'pdf', source, '-o', dest)


def launch_calc(fname):
    """
    Launch calc with open socket communication to start communication with 
    oosheet.
    """
    return headless_cmd(
        '--calc',
        '--accept="socket,'
        'host=localhost,port=2002;urp;StarOffice.ServiceManager"',
        fname,
    )
